@@ -5,6 +5,7 @@ from typing import List, Optional
 
 from . import llama_cpp
 
+
 class Llama:
     def __init__(
         self,
@@ -16,10 +17,11 @@ class Llama:
         logits_all: bool = False,
         vocab_only: bool = False,
         n_threads: Optional[int] = None,
-        model_name: Optional[str]=None,
     ):
         self.model_path = model_path
-        self.model = model_name or model_path
+
+        self.last_n = 64
+        self.max_chunk_size = 32
 
         self.params = llama_cpp.llama_context_default_params()
         self.params.n_ctx = n_ctx
@@ -58,24 +60,42 @@ class Llama:
             stop = [s.encode("utf-8") for s in stop]
 
         prompt_tokens = llama_cpp.llama_tokenize(
-            self.ctx, prompt.encode("utf-8"), self.tokens, self.params.n_ctx, True
+            self.ctx,
+            prompt.encode("utf-8"),
+            self.tokens,
+            llama_cpp.llama_n_ctx(self.ctx),
+            True,
         )
 
         if prompt_tokens + max_tokens > self.params.n_ctx:
             raise ValueError(
-                f"Requested tokens exceed context window of {self.params.n_ctx}"
+                f"Requested tokens exceed context window of {llama_cpp.llama_n_ctx(self.ctx)}"
             )
 
-        for i in range(prompt_tokens):
-            llama_cpp.llama_eval(
-                self.ctx, (llama_cpp.c_int * 1)(self.tokens[i]), 1, i, self.n_threads
+        # Process prompt in chunks to avoid running out of memory
+        for i in range(0, prompt_tokens, self.max_chunk_size):
+            chunk = self.tokens[i : min(prompt_tokens, i + self.max_chunk_size)]
+            rc = llama_cpp.llama_eval(
+                self.ctx,
+                (llama_cpp.llama_token * len(chunk))(*chunk),
+                len(chunk),
+                max(0, i - 1),
+                self.n_threads,
             )
+            if rc != 0:
+                raise RuntimeError(f"Failed to evaluate prompt: {rc}")
 
         for i in range(max_tokens):
+            tokens_seen = prompt_tokens + completion_tokens
+            last_n_tokens = [0] * max(0, self.last_n - tokens_seen) + [
+                self.tokens[j]
+                for j in range(max(tokens_seen - self.last_n, 0), tokens_seen)
+            ]
+
             token = llama_cpp.llama_sample_top_p_top_k(
                 self.ctx,
-                self.tokens,
-                prompt_tokens + completion_tokens,
+                (llama_cpp.llama_token * len(last_n_tokens))(*last_n_tokens),
+                len(last_n_tokens),
                 top_k=top_k,
                 top_p=top_p,
                 temp=temperature,
@@ -84,7 +104,6 @@ class Llama:
             if token == llama_cpp.llama_token_eos():
                 finish_reason = "stop"
                 break
-            # text += llama_cpp.llama_token_to_str(self.ctx, token).decode("utf-8")
             text += llama_cpp.llama_token_to_str(self.ctx, token)
             self.tokens[prompt_tokens + i] = token
             completion_tokens += 1
@@ -98,7 +117,7 @@ class Llama:
 
             llama_cpp.llama_eval(
                 self.ctx,
-                (llama_cpp.c_int * 1)(self.tokens[prompt_tokens + i]),
+                (llama_cpp.llama_token * 1)(self.tokens[prompt_tokens + i]),
                 1,
                 prompt_tokens + completion_tokens,
                 self.n_threads,
@@ -117,12 +136,11 @@ class Llama:
                 self.ctx,
             )[:logprobs]
 
-
         return {
-            "id": f"cmpl-{str(uuid.uuid4())}", # Likely to change
+            "id": f"cmpl-{str(uuid.uuid4())}",  # Likely to change
             "object": "text_completion",
             "created": int(time.time()),
-            "model": self.model, # Likely to change
+            "model": self.model_path,
             "choices": [
                 {
                     "text": text,
@@ -140,5 +158,3 @@ class Llama:
 
     def __del__(self):
         llama_cpp.llama_free(self.ctx)
-
-
